# enables returns of custom types
from __future__ import annotations

from . TourReporter import TourReporter

"""
Knightstour objects
"""
class Chessboard(object):

	def __init__(self, x_size: int, y_size: int, reporter: TourReporter):
		self._reporter = reporter
		self._x_size = x_size
		self._y_size = y_size
		self._tour_limit = -1

		self._board = []
		for i in range(self.y_size):
			self._board.append([-1]*self.x_size)

		self._possible_changes = [(2,1),\
		(1,2),\
		(-1,2),\
		(-2,1),\
		(-2, -1),\
		(-1, -2),\
		(1,-2),\
		(2,-1)\
		]

	@property
	def tour_limit(self):
		return self._tour_limit
	
	@tour_limit.setter
	def tour_limit(self, l):
		"""
		If you want to force a smaller tour
		"""
		self._tour_limit = l



	@property
	def reporter(self):
		return self._reporter
	

	@property
	def tour_length(self):
		"""
		The length of a successful tour
		"""

		if (self.tour_limit != -1):
			return self.tour_limit 

		return self.x_size * self.y_size

	@property
	def x_size(self):
		return self._x_size

	@property
	def y_size(self):
		return self._y_size
	

	@property
	def board(self):
		return self._board
	


	@property
	def possible_changes(self):
		return self._possible_changes


	
	def pop_move(self, pos:tuple(int, int)):
		"""
		Remove the last move
		"""
		self.board[pos[0]][pos[1]] = -1

	def add_move(self, pos:tuple(int, int), move_count:int):
		"""
		Add a move to the array of used moves
		"""

		self.board[pos[0]][pos[1]] = move_count
		self.reporter.increment_moves()

	def solve_board(self) -> bool:
		
		for i in range(self.x_size):
			for ii in range(self.y_size):
				pos = (i, ii)
				self.solve_tour(pos, 1)


	def solve_tour(self, start_position:tuple(int,int), move_count:int) -> bool:
		self.add_move(start_position, move_count)

		if (move_count >= self.tour_length):
			print ("!!! WE HAVE A WINNER !!!")
			self.reporter.report_tour(self.board)
			return True


		#if (move_count > 62):
		#	print("{}".format(self.board))


		#if (self.current_depth >= 15 ):
		#	print ("!!!!! Quadrant Solved !!! {}".format(self.moves))


		for i in range(len(self.possible_changes)):
			pos = self.new_position(start_position, self.possible_changes[i])
			if self.is_legal(pos):
				#print("CHECKING MOVE: depth:{}, {},{} -> {},{}".format(move_count, start_position[0],start_position[1],  pos[0], pos[1]))

				#self.add_move(pos, move_count)
				self.solve_tour(start_position=pos, move_count=move_count+1)
				self.pop_move(pos)

		return False



	def is_legal(self, position:tuple(int,int)) -> bool:
		"""
		return true if the move is open or legal
		"""

		# out of bounds
		if (position[0] < 0 \
			or position[1] < 0 \
			or position[0] >= self.x_size \
			or position[1] >= self.y_size):

			#print("NOT LEGAL:     out of bounds:{},{}".format(position[0], position[1]))
			return False



		# already used
		if self.board[position[0]][position[1]] != -1:
			#print("NOT LEGAL:     already visited position:{},{}".format(position[0], position[1]))
			return False



		return True







	def new_position(self, start:tuple(int,int), change:tuple[int,int]) -> tuple[int,int]:
		"""
		Return a new position.
		Note, this is not guaranteed to be a legal move.

		Paramteres:
			start: starting position
			change: the change to apply

		Returns:
			tuple: a new position
		"""

		pos = (start[0] + change[0], start[1] + change[1])
		return pos











# enables returns of custom types
from __future__ import annotations

"""
Object to collect and report on all the knightstours.
"""
class TourReporter(object):

	def __init__(self, logging=True):
		self._tours_count = 0
		self._moves = 0
		self._logging = logging

		if self.logging:
			self._output = open("/ext/output_data/tours.log", "w+", 1)

	@property
	def logging(self):
		return self._logging
	

	@property
	def tours_count(self):
		"""
		Returns:
			count (int): how many tours have been reported
		"""

		return self._tours_count
	

	@property
	def moves(self):
		"""
		Each time a move is cheked, increment this count.

		Returns:
			move (int): the number of permutations.
		"""

		return self._moves
	
	def increment_moves(self):
		"""
		Increment the move count.
		"""

		self._moves = self.moves + 1

		if ((self.moves % 10000) == 0):
			print("=== Have checked another 10,000 moves: {}".format(self.moves))

		#	if (self.logging):
		#		self._output.write("=== Have checked another 10,000 moves: {}\n".format(self.moves))


	def report_tour(self, newTour: [int][int]):
		"""
		Report a new tour that has been found.

		Parameters:
			newTour: a list of tuples for each move in the tour.

		"""
		self._tours_count = self._tours_count + 1


		print("Reporting tour: {}".format(self._tours_count))

		if self.logging:
			self._output.write("Knights tour: {0}\n".format(self._tours_count))
			self._output.write("--------\n")

			for row in newTour:
				self._output.write("    {0}\n".format(row[0:]))
			self._output.write("\n")




N = 8
CC = 0

def is_valid(i, j, sol):
  if (i>=1 and i<=N and j>=1 and j<=N):
    if (sol[i][j]==-1):
      return True
  return False

def knight_tour(sol, i, j, step_count, x_move, y_move):
  global CC

  if (step_count == N*N):
    return True

  for k in range(0, 8):
    next_i = i+x_move[k]
    next_j = j+y_move[k]

    if(is_valid(next_i, next_j, sol)):
      sol[next_i][next_j] = step_count
      if (knight_tour(sol, next_i, next_j, step_count+1, x_move, y_move)):
        return True
      sol[next_i][next_j] = -1; # backtracking
  return False

def start_knight_tour():
  sol = []

  for i in range(0, N+1):
    a = [0]+([-1]*N)
    sol.append(a)

  x_move = [2, 1, -1, -2, -2, -1, 1, 2]
  y_move = [1, 2, 2, 1, -1, -2, -2, -1]

  sol[1][1] = 0 # placing knight at cell(1, 1)

  if (knight_tour(sol, 1, 1, 1, x_move, y_move)):
    for i in range(1, N+1):
      print(sol[i][1:])
    return True
  return False

if __name__ == '__main__':
  print(start_knight_tour())


import sys
sys.path.append("./modules/")

from knightstour.Chessboard import Chessboard
from knightstour.TourReporter import TourReporter

tours = TourReporter()
board = Chessboard(x_size=8, y_size=8, reporter=tours)
#board.tour_limit = 64
board.solve_board()

print("moves checked: {0}".format(tours.moves))
print("we found {0} tours".format(tours.tours_count))





# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['knightstour', 'knightstour.modules.knightstour']

package_data = \
{'': ['*'], 'knightstour': ['modules/*']}

setup_kwargs = {
    'name': 'knightstour',
    'version': '0.1.1',
    'description': 'A simplistic implementation of the knights tour',
    'long_description': '# Knights Tour 0.1.x\n\nA very simple implementation of the knights tour. Intended to create a lot of data work with a little bit of code.\n\n## Installation\n\nUse poetry to install package\n\n```bash\npoetry install knightstour-lucasmcgregor\n```\n\n## Usage\n\n```python\nfrom knightstour.Chessboard import Chessboard\nfrom knightstour.TourReporter import TourReporter\n\ntours = TourReporter()\nboard = Chessboard(x_size=8, y_size=8, reporter=tours)\nboard.solve_board()\n\nprint("moves checked: {0}".format(tours.moves))\nprint("we found {0} tours".format(tours.tours_count))\n```\n\n## License\n[MIT](https://choosealicense.com/licenses/mit/)',
    'author': 'Lucas McGregor',
    'author_email': 'lucasmcgregor@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
